@@ -3,7 +3,7 @@
  *  XVID MPEG-4 VIDEO CODEC
  *  - Image management functions -
  *
- *  Copyright(C) 2001-2004 Peter Ross <pross@xvid.org>
+ *  Copyright(C) 2001-2010 Peter Ross <pross@xvid.org>
  *
  *  This program is free software ; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@
  *  along with this program ; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
  *
- * $Id: image.c,v 1.43 2008/11/28 10:58:07 Isibaar Exp $
+ * $Id: image.c,v 1.46.2.1 2010/12/30 11:46:58 Isibaar Exp $
  *
  ****************************************************************************/
 
@@ -149,7 +149,7 @@ image_setedges(IMAGE * image,
 
 	/* According to the Standard Clause 7.6.4, padding is done starting at 16
 	 * pixel width and height multiples. This was not respected in old xvids */
-	if (bs_version == 0 || bs_version >= SETEDGES_BUG_BEFORE) {
+	if (bs_version >= SETEDGES_BUG_BEFORE) {
 		width  = (width+15)&~15;
 		height = (height+15)&~15;
 	}
@@ -390,15 +390,20 @@ safe_packed_conv(uint8_t * x_ptr, int x_stride,
 {
 	int width_opt, width_c, height_opt;
 
-    if (width==1 || height==1) return; /* forget about it */
+    if (width<0 || width==1 || height==1) return; /* forget about it */
 
 	if (func_opt != func_c && x_stride < size*((width+15)/16)*16)
 	{
 		width_opt = width & (~15);
 		width_c = (width - width_opt) & (~1);
 	}
-	else
+	else if (func_opt != func_c && !(width&1) && (size==3))
 	{
+        /* MMX reads 4 bytes per pixel for RGB/BGR */
+        width_opt = width - 2;
+        width_c = 2;
+    }
+    else {
         /* Enforce the width to be divisable by two. */
 		width_opt = width & (~1);
 		width_c = 0;
@@ -857,6 +862,98 @@ long plane_sse(uint8_t *orig,
 	return (sse);
 }
 
+void image_block_variance(IMAGE * orig_image,
+				          uint16_t stride, 
+				          MACROBLOCK *mbs,
+				          uint16_t mb_width, 
+				          uint16_t mb_height)
+{
+	DECLARE_ALIGNED_MATRIX(sums, 1, 4, uint16_t, CACHE_LINE);
+	DECLARE_ALIGNED_MATRIX(squares, 1, 4, uint32_t, CACHE_LINE);
+	
+	int x, y, i, j;
+	uint8_t *orig_y = orig_image->y;
+	uint8_t *orig_u = orig_image->u;
+	uint8_t *orig_v = orig_image->v;
+
+	for (y = 0; y < mb_height; y++) {
+		for (x = 0; x < mb_width; x++) {
+			MACROBLOCK *pMB = &mbs[x + y * mb_width];
+			uint32_t var4[4];
+			uint32_t sum = 0, square = 0;
+
+			/* y-blocks */
+			for (j = 0; j < 2; j++) {
+				for (i = 0; i < 2; i++) {
+					int lsum = blocksum8(orig_y + ((y<<4) + (j<<3))*stride + (x<<4) + (i<<3), 
+										 stride, sums, squares);
+					int lsquare = (squares[0] + squares[1] + squares[2] + squares[3])<<6;
+
+					sum += lsum;
+					square += lsquare;
+
+					var4[0] = (squares[0]<<4) - sums[0]*sums[0];
+					var4[1] = (squares[1]<<4) - sums[1]*sums[1];
+					var4[2] = (squares[2]<<4) - sums[2]*sums[2];
+					var4[3] = (squares[3]<<4) - sums[3]*sums[3];
+
+					pMB->rel_var8[j*2 + i] = lsquare - lsum*lsum;
+					if (pMB->rel_var8[j*2 + i])
+						pMB->rel_var8[j*2 + i] = ((var4[0] + var4[1] + var4[2] + var4[3])<<8) / 
+												 pMB->rel_var8[j*2 + i]; /* 4*(Var(Di)/Var(D)) */
+					else 
+						pMB->rel_var8[j*2 + i] = 64;
+				}
+			}
+
+			/* u */
+			{
+				int lsum = blocksum8(orig_u + (y<<3)*(stride>>1) + (x<<3), 
+									 stride, sums, squares);
+				int lsquare = (squares[0] + squares[1] + squares[2] + squares[3])<<6;
+
+				sum += lsum;
+				square += lsquare;
+
+				var4[0] = (squares[0]<<4) - sums[0]*sums[0];
+				var4[1] = (squares[1]<<4) - sums[1]*sums[1];
+				var4[2] = (squares[2]<<4) - sums[2]*sums[2];
+				var4[3] = (squares[3]<<4) - sums[3]*sums[3];
+
+				pMB->rel_var8[4] = lsquare - lsum*lsum;
+				if (pMB->rel_var8[4])
+					pMB->rel_var8[4] = ((var4[0] + var4[1] + var4[2] + var4[3])<<8) / 
+										 pMB->rel_var8[4]; /* 4*(Var(Di)/Var(D)) */
+				else 
+					pMB->rel_var8[4] = 64;
+			}
+
+			/* v */
+			{
+				int lsum = blocksum8(orig_v + (y<<3)*(stride>>1) + (x<<3), 
+									 stride, sums, squares);
+				int lsquare = (squares[0] + squares[1] + squares[2] + squares[3])<<6;
+
+				sum += lsum;
+				square += lsquare;
+
+				var4[0] = (squares[0]<<4) - sums[0]*sums[0];
+				var4[1] = (squares[1]<<4) - sums[1]*sums[1];
+				var4[2] = (squares[2]<<4) - sums[2]*sums[2];
+				var4[3] = (squares[3]<<4) - sums[3]*sums[3];
+
+				pMB->rel_var8[5] = lsquare - lsum*lsum;
+				if (pMB->rel_var8[5])
+					pMB->rel_var8[5] = ((var4[0] + var4[1] + var4[2] + var4[3])<<8) / 
+										 pMB->rel_var8[5]; /* 4*(Var(Di)/Var(D)) */
+				else 
+					pMB->rel_var8[5] = 64;
+			}
+
+		}
+	}
+}
+
 #if 0
 
 #include <stdio.h>
